open Ast;;
open Printer;;
open Typer;;
open Ast_typer;;

type interpreting_error =
  | UnexpectedType of java_type * java_type
  | UnsupportedOperator;;

exception Interpreter_break;;
exception Interpreter_continue;;
exception Interpreter_return;;
exception Interpreting_error of interpreting_error;;

type inter_env = {
  trace : bool;
  mutable inter_main : typed_method_def option;
  classes : (string, typed_class) Hashtbl.t;
  mutable block_current : int;
  block_stack : int array;   (* stack pointer for a given block *)
  mutable stack_pointer : int;
  local_stack : typed_data array;
  
  (* quick stack for the arguments *)
  mutable args_current : int;
  args_stack : typed_data array;
  
  (* returned values *)
  mutable return_val : typed_data;
};;

let print_interpreting_error oc e =
  match e with
  | UnexpectedType(lt1, lt2) ->
      output_string oc ("Got type " ^ (string_of_java_type lt1)
          ^ ", but was expecting "
          ^ (string_of_java_type lt2) ^ "\n");
      failwith "dynamic_type_check"
  | UnsupportedOperator -> failwith "unsupported operator";;

(* convert from full precision to a given number of bits... *)
let bit_precision i java_typ =
  let bit_prec bits =
    let n = (i land ((1 lsl bits) - 1)) and max = (1 lsl (bits - 1))
    in if n < max then n else (1 lsl bits) - n
  in match java_typ with
    | `byte -> bit_prec 8
    | `short | `char -> bit_prec 16
    | _ -> i;;

let get_type exp =
  match exp with
  | Void -> `void
  | Bool _ -> `boolean
  | Int _ -> `int
  | Float _ -> `float
  | String _ -> `string
  | Object (c, _) -> `java_object([c.tc_name])
  | Null -> `nil;;

let get_int exp =
  match exp with
  | Int(i) -> i
  | _ -> raise (Interpreting_error( UnexpectedType (get_type exp, `int) ));;

let get_float exp =
  match exp with
  | Float(f) -> f
  | _ -> raise (Interpreting_error( UnexpectedType (get_type exp, `float) ));;

let get_bool exp =
  match exp with
  | Bool(b) -> b
  | _ -> raise (Interpreting_error( UnexpectedType (get_type exp, `boolean) ));;

let get_string exp =
  match exp with
  | String(s) -> s
  | _ -> raise (Interpreting_error( UnexpectedType (get_type exp, `string) ));;

let get_object_data exp =
  match exp with
    | Object(_, d) -> d
    | _ -> raise (Interpreting_error( UnexpectedType (get_type exp, `java_object([]) )));;

let reduce_type t =
  match t with
  | `byte | `short | `char -> `int
  | _ -> t;;

let read_object obj =
  match obj with
  | Object(c, data) -> c, data
  | Null _ -> failwith "Trying to access null object"
  | _ -> failwith ("ERROR : NOT AN OBJECT, got type " ^ (string_of_java_type (get_type obj)));;

(* READ/WRITE variables *)
let store_in_stack env block_id var_id e =
  let stack_id = (Array.get env.block_stack (env.block_current - block_id)) + var_id
  in
    Array.set env.local_stack stack_id e;;

let read_in_stack env block_id var_id =
  let stack_id = (Array.get env.block_stack (env.block_current - block_id)) + var_id
  in
    (Array.get env.local_stack stack_id);;

let store_in_params env var_id e =
  env.args_stack.(env.args_current - var_id - 1) <- e;;

let read_in_params env var_id =
  env.args_stack.(env.args_current - var_id - 1);;

let store_in_fields env var_id e =
  let obj = env.args_stack.(env.args_current)
  in let c, data = read_object obj
  in data.(var_id) <- e;;

let read_in_fields env var_id =
  let obj = env.args_stack.(env.args_current)
  in let c, data = read_object obj
  in data.(var_id);;

(* CLASS MANAGEMENT *)
let rec get_nth_superclass c n =
  if n = 0 then c else
    match c.tc_father with
    | None -> failwith "shouldn't be here..."
    | Some(p) -> get_nth_superclass p (n - 1);;

let get_superclass c =
  get_nth_superclass c 1;;

let get_class env class_name =
  try
    Hashtbl.find env.classes class_name
  with
  | Not_found -> failwith ("No class named " ^ class_name ^ "\n%!");;

let get_method c id =
  let (_, _, s, m) = c.tc_methods.(id)
  in m;;

let rec get_constructor c super_id id =
  if super_id = 0 then c.tc_cons.(id)
  else get_constructor (get_nth_superclass c super_id) 0 id;;

let rec get_parent_class_name c id =
  if (id = 0) then c.tc_name
  else get_parent_class_name (get_superclass c) (id - 1);;

let rec is_subclass c name =
  if (String.compare c.tc_name name) = 0 then true
  else match c.tc_father with
    | None -> false
    | Some(p) -> is_subclass p name;;

let nyi s = Printf.printf "%s not yet implemented\n%!" s;;

(* EVAL EXPRESSIONS *)
let rec eval_exp env exp =
  let rec eval_e = eval_exp env
  in let rec eval_cast cast e =
    begin
      match cast with
      | Ast_typer.Float_of_int -> Float(float_of_int (get_int e))
      | Ast_typer.Int_of_float -> Int(int_of_float (get_float e))
      | Ast_typer.Int_of_int -> e
      | Ast_typer.String_of_int -> String(string_of_int (get_int e))
      | Ast_typer.String_of_float -> String(string_of_float (get_float e))
      | Ast_typer.String_of_bool -> String(string_of_bool (get_bool e))
      | Ast_typer.String_of_object -> String("object")
    end
  and eval_unop t op e =
    begin
      let unop_int add is_pre =
        match e with
        | t, Evar (Alocal (jt, block_id, var_id))
        -> begin
              let var = (read_in_stack env block_id var_id)
              in let v = (get_int var) + add
              in let new_var = Int(v)
              in store_in_stack env block_id var_id new_var ; if (is_pre) then new_var else var
            end
        | _ -> failwith "++ and -- only work on variables..."
      and unop_float add is_pre =
        match e with
        | t, Evar (Alocal (jt, block_id, var_id))
        -> begin
              let var = (read_in_stack env block_id var_id)
              in let v = (get_float var) +. add
              in let new_var = Float(v)
              in store_in_stack env block_id var_id new_var ; if (is_pre) then new_var else var
            end
        | _ -> failwith "++ and -- only work on variables..."
      
      in match (reduce_type t), op with
        | `int, Oneg -> Int(- (get_int (eval_e e)))
        | `float, Oneg -> Float(-.(get_float (eval_e e)))
        | `boolean, Onot -> Bool(not (get_bool (eval_e e)))
        | `int, Otild -> Int(lnot (get_int (eval_e e)))
        | `int, Opreinc -> unop_int 1 true
        | `int, Opredec -> unop_int (- 1) true
        | `int, Opostinc -> unop_int 1 false
        | `int, Opostdec -> unop_int (- 1) false
        | `float, Opreinc -> unop_float 1.0 true
        | `float, Opredec -> unop_float (- 1.0) true
        | `float, Opostinc -> unop_float 1.0 false
        | `float, Opostdec -> unop_float (- 1.0) false
        | _ -> Printf.printf "Unsupported unop\n%!" ; raise (Interpreting_error(UnsupportedOperator))
    end
  and eval_binop t op e1 e2 =
    begin
      (* warning : comparaison paraisseuse fail ! *)
      let res1 = eval_e e1 and res2 = eval_e e2 in
        match ((reduce_type t), op) with
        | `int, Oplus -> Int((get_int res1) + (get_int res2))
        | `float, Oplus -> Float((get_float res1) +. (get_float res2))
        | `string, Oplus -> String((get_string res1) ^ (get_string res2))
        | `int, Odiv -> Int((get_int res1) / (get_int res2))
        | `float, Odiv -> Float((get_float res1) /. (get_float res2))
        | `int, Ominus -> Int((get_int res1) - (get_int res2))
        | `float, Ominus -> Float((get_float res1) -. (get_float res2))
        | `int, Omul -> Int((get_int res1) * (get_int res2))
        | `float, Omul -> Float((get_float res1) *. (get_float res2))
        | `int, Omod -> Int((get_int res1) mod (get_int res2))
        | `boolean, Oandand -> Bool((get_bool res1) && (get_bool res2))
        | `boolean, Ooror -> Bool((get_bool res1) || (get_bool res2))
        | _ -> Printf.printf "Unsupported binop\n%!" ; raise (Interpreting_error(UnsupportedOperator))
    end
  and eval_bincomp op e1 e2 =
    begin
      let res1 = eval_e e1 and res2 = eval_e e2 in
      let t = (reduce_type (fst e1)) in
      let rec comp op =
        match (t, op) with
        | `int, Oequal -> Bool((get_int res1) = (get_int res2))
        | `float, Oequal -> Bool((get_float res1) = (get_float res2))
        | `boolean, Oequal -> Bool((get_bool res1) = (get_bool res2))
        | `string, Oequal -> Bool((get_string res1) = (get_string res2))
        | `java_object(_), Oequal -> 
          begin
            match res1, res2 with
              | Null, Null -> Bool(true)
              | Object(_, d1), Object(_, d2) -> Bool(d1 == d2)
              | _ -> Bool(false)
          end
        | _, Onotequal -> (
              match (comp Oequal) with
              | Bool(b) -> Bool(not b)
              | _ -> assert false )
        | `int, Osup -> Bool((get_int res1) > (get_int res2))
        | `float, Osup -> Bool((get_float res1) > (get_float res2))
        | `int, Oinf -> Bool((get_int res1) < (get_int res2))
        | `float, Oinf -> Bool((get_float res1) < (get_float res2))
        | `int, Osupequal -> Bool((get_int res1) >= (get_int res2))
        | `float, Osupequal -> Bool((get_float res1) >= (get_float res2))
        | `int, Oinfequal -> Bool((get_int res1) <= (get_int res2))
        | `float, Oinfequal -> Bool((get_float res1) <= (get_float res2))
        | _ -> Printf.printf "Unsupported bincomp\n%!" ; raise (Interpreting_error(UnsupportedOperator))
      in comp op
    end
  
  and eval_ter cond e1 e2 =
    if (get_bool (eval_e cond)) then eval_e e1 else eval_e e2
  
  and eval_affect var e1 =
    begin
      (match var with
        | Alocal (jt, block_id, var_id) -> store_in_stack env block_id var_id e1
        | Aparam (jt, var_id) -> store_in_params env var_id e1
        | Afield (jt, block_id, var_id) -> store_in_fields env var_id e1
        | _ -> failwith "Unmanaged affectation...\n%!"
      ) ;
      e1
    end
  
  and eval_var var =
    match var with
    | Alocal (jt, block_id, var_id) -> read_in_stack env block_id var_id
    | Aparam (jt, var_id) -> read_in_params env var_id
    | Afield (jt, block_id, var_id) -> read_in_fields env var_id
    | _ -> failwith "Unmanaged var\n%!"
  
  and eval_new_object ct id params =
    let c = get_class env (List.hd ct)
    in let a = Array.make c.tc_field_count Void
    in let obj = Object(c, a)
    in let rec call_constructor obj super_id id params =
      begin
        if env.trace then
          begin
            (if super_id > 0 then Printf.printf "  " );
            let c = get_nth_superclass c super_id
            in Printf.printf "TRACE %s( %s )\n%!" c.tc_name (string_of_java_types (List.map fst params))
          end
      end;
      let cons = get_constructor c super_id id
      in (* other constructor to call ?? *)
        begin
          match cons.typed_constructor_call with
          | Cnone -> ()
          | Cthis(id, p) -> call_constructor obj super_id id p
          | Csuper(id, p) -> call_constructor obj (super_id + 1) id p
        end ;
        let _ = do_call obj params cons.typed_constructor_body in ()
    in
      call_constructor obj 0 id params ;
      obj
  
  and do_call obj p body =
    match body with
    | None -> Void
    | Some b ->
        begin
          (* stack arguments and object *)
          let params = List.map eval_e p
          in let size = List.length params and n = ref 0
          in let add_to_stack i =
            incr n;
            env.args_stack.(env.args_current - !n) <- i
          in env.args_current <- env.args_current + size + 1;
            (* add object to stack *)
            env.args_stack.(env.args_current) <- obj;
            (* add parameters to stack *)
            List.iter add_to_stack params ;
            
            let stack_pointer = env.stack_pointer
            and block_current = env.block_current
            in
              begin
                try
                  eval_instr env b;
                with
                | Interpreter_return -> ()
              end ;
              
              (* restore stack state *)
              env.stack_pointer <- stack_pointer;
              env.block_current <- block_current;
              
              (* unstack... *)
              env.args_current <- env.args_current - size - 1;
              env.return_val
        end
  and eval_call obj id params =
    let c, da = read_object obj
    in let (parent_id, _, name, met) = c.tc_methods.(id)
    in (if env.trace then Printf.printf "TRACE %s.%s\n%!" (get_parent_class_name c parent_id) name);
      do_call obj params met.typed_method_body
  
  and eval_expr_call te s id tel =
    eval_call (eval_e te) id tel
  
  and get_current_obj () =
    env.args_stack.(env.args_current)
  
  and eval_instanceof e t =
    let te = eval_e e
    and superclass = match t with
      | `java_object (o) -> List.hd o
      | _ -> failwith "NOT A CLASS"
    
    in let c, _ = read_object te
    in Bool(is_subclass c superclass)
  
  and eval_field_access e id =
    let v = eval_e e
    in let _, vars = read_object v
    in vars.(id)
  
  and nyi str = failwith (str ^ " not yet implemented")
  in match exp with
    | t, Ast_typer.Eint (i) -> Int(i)(*Int(bit_precision i t)*)
    | _, Ast_typer.Efloat(f) -> Float(f)
    | _, Ast_typer.Estring(s) -> String(s)
    | _, Ast_typer.Ebool(b) -> Bool(b)
    | _, Ast_typer.Ecast( cast, e ) -> let exp = eval_e e in eval_cast cast exp
    | t, Ast_typer.Eunary (op, e) -> eval_unop t op e
    | t, Ast_typer.Ebinary( op, e1, e2 ) -> eval_binop t op e1 e2
    | _, Ast_typer.Ebincomp( op, e1, e2) -> eval_bincomp op e1 e2
    | _, Ast_typer.Eternary(_, cond, e1, e2) -> eval_ter cond e1 e2
    | _, Ast_typer.Eaffect(var, e1) -> eval_affect var (eval_e e1)
    | _, Ast_typer.Evar(aff) -> eval_var aff
    | t, Ast_typer.Enew_object(ct, id, el) -> eval_new_object ct id el
    | _, Ast_typer.Enull -> Null
    | _, Ast_typer.Ecall (s, id, tel) -> eval_call (get_current_obj ()) id tel
    | _, Ast_typer.Eexpr_call (te, s, id, tel) -> eval_expr_call te s id tel
    | _, Ast_typer.Einstanceof (e, t) -> eval_instanceof e t
    | _, Ast_typer.Eexpr_cast (_, _) -> nyi "Eexpr_cast"
    | _, Ast_typer.Etype_cast (_, _, _) -> nyi "Etype_cast"
    | _, Ast_typer.Efield_access (e, _, id) -> eval_field_access e id
    | _, Ast_typer.Enew_array (_, _, _) -> nyi "Enew_array"
    | _, Ast_typer.Earray_access (_, _) -> nyi "Earray_access"
    | _, Ast_typer.Ethis -> get_current_obj ()
    | _, Ast_typer.Esuper -> nyi "Esuper"
(* | _ -> Printf.printf "Unsupported exp\n%!" ; raise                      *)
(* (Interpreting_error(UnsupportedOperator))                               *)

and eval_instr env i =
  let eval_i = eval_instr env and eval_e = eval_exp env
  in let rec eval_block decl instr =
    let n = List.length decl in
      begin
        (* PUSH *)
        env.block_current <- env.block_current + 1;
        Array.set env.block_stack env.block_current env.stack_pointer;
        List.iter eval_i decl ;
        
        (* incr env.block_current; *)
        
        List.iter eval_i instr;
        
        (* POP *)
        env.block_current <- env.block_current - 1;
        env.stack_pointer <- env.stack_pointer - n;
      end
  and eval_decl e =
    match e with
    | Simple_init(exp)
    -> let te = eval_e exp
        in begin
            (* PUSH *)
            Array.set env.local_stack env.stack_pointer te ;
            env.stack_pointer <- env.stack_pointer + 1;
          end
    | _ -> failwith "unmanaged declaration"
  
  and eval_while exp instr =
    let block = env.block_current and sp = env.stack_pointer
    in let restore_env = env.block_current <- block ; env.stack_pointer <- sp
    in try
        while (get_bool (eval_e exp)) do eval_i instr done
      with
      | Interpreter_break -> env.block_current <- block ; env.stack_pointer <- sp
      | Interpreter_continue -> restore_env ; eval_while exp instr
  and eval_do_while instr exp =
    begin
      let block = env.block_current and sp = env.stack_pointer
      in let restore_env = env.block_current <- block ; env.stack_pointer <- sp
      in try eval_i instr ; eval_while exp instr
        with
        | Interpreter_break -> restore_env
        | Interpreter_continue -> restore_env ; eval_while exp instr
    end
  
  and eval_return exp =
    begin
      match exp with
      | None -> env.return_val <- Void
      | Some(e) -> env.return_val <- eval_e e
    end ; raise (Interpreter_return)
  
  and eval_ifthenelse c e1 e2 =
    match eval_e c with
    | Bool(true) -> eval_i e1
    | Bool(false) -> eval_i e2
    | _ -> failwith "condition is not a boolean..."
  
  and eval_println s =
    let eval_str = (match (eval_e s) with
        | String str -> str
        | _ -> failwith "unexpected type instead of string")
    in Printf.printf "%s\n%!" eval_str
  
  in match i with
    | Ast_typer.Ilabeled _ -> nyi "instruction label..."
    | Ast_typer.Iexpr e -> let _ = eval_e e in ()
    | Ast_typer.Iblock (decl, instr) -> eval_block decl instr
    | Ast_typer.Idecl (_, _, e) -> eval_decl e
    | Ast_typer.Iifthenelse (cond, e1, e2) -> eval_ifthenelse cond e1 e2
    | Ast_typer.Iswitch _ -> nyi "instruction switch..."
    | Ast_typer.Iwhile (exp, instr) -> eval_while exp instr
    | Ast_typer.Idowhile (instr, exp) -> eval_do_while instr exp
    | Ast_typer.Ibreak _ -> raise Interpreter_break
    | Ast_typer.Icontinue _ -> raise Interpreter_continue
    | Ast_typer.Ireturn e -> eval_return e
    | Ast_typer.Ithrow _ -> nyi "instruction throw..."
    | Ast_typer.Isynchronized _ -> nyi "instruction sync..."
    | Ast_typer.Itry _ -> nyi "instruction try..."
    | Ast_typer.Inop -> ()
    | Ast_typer.Iprintln s -> eval_println s

and eval_class_tree env obj =
  let rec rec_ect obj =
    begin
      (* Search for main method *)
      let find_main m =
        begin
          let (_, _, s, tm) = m
          in if (String.compare (String.sub s 0 5) "main(") = 0 then env.inter_main <- Some(tm)
        end in
        Array.iter find_main obj.tc_methods ;
        Hashtbl.add env.classes obj.tc_name obj ;
        List.iter rec_ect obj.tc_children;
    end
  in rec_ect obj ;
    match env.inter_main with
    | None -> failwith "No main method found, exiting !\n%!"
    | Some(m) -> (* Run everything ! *)
        begin
          match m with
          | { typed_method_body = None } -> failwith "empty body of main method"
          | { typed_method_body = Some b } -> eval_instr env b;
        end

and eval_compil_unit {
      typed_package = typed_package;
      typed_imports = typed_imports;
      typed_class = typed_class;
      typed_interf_decls = typed_interf_decls;
    } =
  try
    let size = 1024 * 1024
    in let env = {
      trace = false;
      inter_main = None;
      classes = Hashtbl.create 256;
      block_current = - 1;
      block_stack = Array.make size 0;
      stack_pointer = 0;
      local_stack = Array.make (size) Void;
      args_current = 0;
      args_stack = Array.make size Void;
      return_val = Void;
    } in
      eval_class_tree env typed_class;
  with Interpreting_error(ie) -> print_interpreting_error stderr ie
;;
