open Ast;;
open Ast_typer;;

exception MultipleMethods of string list;;
type overloading_algorithm =
  | First
  | Java_1_2
  | Java_1_5
  | Multiply
  | Perfect;;

let algorithm = Java_1_2;;

let get_algorithm_name = function
  | First -> "First Match"
  | Java_1_2 -> "Java 1.2"
  | Java_1_5 -> "Java 1.5"
  | Multiply -> "Multiplication"
  | Perfect -> "Perfect Match";;

let get_class class_name obj =
  let rec r_get_class obj =
    let rec get_class_in_list l =
      match l with
      | [] -> None
      | h:: t ->
          begin
            let c = r_get_class h in match c with
              | Some(cl) -> c
              | None -> get_class_in_list t
          end
    in if (String.compare obj.ic_name class_name) = 0 then Some(obj)
      else get_class_in_list obj.ic_children
  
  in match r_get_class obj with
    | Some (c) -> c
    | None -> raise(Not_found);;

(* test if a class is subclass of other class *)
let get_class_dist classes subclass superclass =
  let rec get_dist obj n =
    if (String.compare superclass obj.ic_name) = 0 then n
    else
      match obj.ic_father with
      | None -> - 1 (* INVALID ! *)
      | Some (p) -> get_dist p (n + 1)
  in get_dist (get_class subclass classes) 0;;

let is_instance_of classes subclass superclass =
  if (get_class_dist classes subclass superclass) = - 1 then false
  else true;;

(* distance between 2 classes *)
let get_types_distance classes t1 t2 =
  let type_flatten t =
    match t with
    | `byte | `char | `short -> `int
    | _ -> t
  in let tf1 = type_flatten t1 and tf2 = type_flatten t2
  in match tf1, tf2 with
    | _, _ when t1 = t2 -> 0
    | `int, `float -> 1
    | `nil, `java_object(o1) -> get_class_dist classes (List.hd o1) "Object"
    | `java_object(o1), `java_object(o2)
    -> get_class_dist classes (List.hd o1) (List.hd o2)
    | _ -> - 1;;  (* invalid distance *)

let are_types_compatible classes t1 t2 =
  (get_types_distance classes t1 t2) <> (- 1);;

let are_all_types_compatible classes l1 l2 =
  List.for_all2 (are_types_compatible classes) l1 l2;;

let get_types_distances obj l1 l2 =
  List.map2 (get_types_distance obj) l1 l2;;

(* OVERLOADING for dummies : just take the first one... *)
let get_first_candidate _ candidates =
  snd (List.hd candidates);;

(* let get_best_candidates list_op candidates = let rec get_smallest l =   *)
(* match l with | [] -> raise Not_found | (d, c)::[] -> let dist = list_op *)
(* d in Printf.printf "dist %i\n" dist ; (dist, c) | (d1, c1):: t -> begin *)
(* let i2 = get_smallest t in let dist1 = list_op d1 and dist2 = fst i2 in *)
(* Printf.printf "dist %i\n" dist1 ; if (dist1 < dist2) then (dist1, c1)   *)
(* else (dist2, (snd i2)) end in let l = snd (get_smallest candidates)     *)

let get_mangled_name current_class (class_id, _, _, mangled_name, _) =
  let rec get_class_name c id = match id, c.ic_father with
    | 0, _ -> c.ic_name
    | n, Some(p) -> get_class_name p (n - 1)
    | _, _ -> assert false
  in (get_class_name current_class class_id) ^ "." ^ mangled_name;;

let get_best_candidate current_class list_op candidates =
  let rec get_smallest l res =
    match l with
    | [] -> res
    | (d1, c1):: t ->
        begin
          let dist = list_op d1
          in let r = match res with
            | [] -> [ (dist, c1) ]
            | (dist2, _):: r2 ->
                begin
                  if (dist < dist2) then [ (dist, c1) ]
                  else if (dist > dist2) then res
                  else (dist, c1):: res  (* Add to current list of overload... *)
                end
          in get_smallest t r
        end
  
  in let l = get_smallest candidates []
  in match l with
    | [] -> raise Not_found
    | [x] -> snd x
    | _ -> let get_string m = get_mangled_name current_class (snd m)
        in raise(MultipleMethods(List.map get_string l));;

(* OVERLOADING add distances between args and parameters (Java style) *)
let get_add_candidate current_class candidates =
  let op l = List.fold_left (+) 0 l
  in get_best_candidate current_class op candidates;;

(* OVERLOADING multiply distances between args and parameters *)
let get_mult_candidate current_class candidates =
  let op l = List.fold_left (fun a b -> a * succ b) 1 l
  in get_best_candidate current_class op candidates;;

(* OVERLOADING only keep a perfect match *)
let rec get_perfect_candidate current_class l =
  match l with
  | [] -> raise Not_found
  | (d1, c1):: t ->
      let dist1 = List.fold_left (+) 0 d1
      in if dist1 = 0 then c1 else get_perfect_candidate current_class t;;

let find_and_map find map l =
  let rec f_a_m l =
    match l with
    | [] -> []
    | h:: t -> if find h then (map h):: (f_a_m t) else f_a_m t
  in f_a_m l;;

let overload verbose algorithm current_class classes name param_types method_list =
  (* (class_id, method_id, met.method_name, mangled_name, param_type_list) *)
  let method_2_call = current_class.ic_name ^ "." ^ name ^ "( " ^ (string_of_java_types param_types) ^ " )";
  and param_count = List.length param_types
  in
  (* candidate matching : same name, same arity, compatible types *)
  let method_match (_, _, met_name, _, met_params) =
    ((List.length met_params) = param_count)
    && (String.compare name met_name) = 0
    && (are_all_types_compatible classes param_types met_params)
  and method_add_dist m =
    let (class_id, _, _, _, met_params) = m
    in let distances = get_types_distances classes param_types met_params
    in match algorithm with
      | Java_1_2 -> (class_id:: distances, m)  (* take class into account*)
      | _ -> (distances, m)
  in
  
  (* get all possible methods *)
  let candidates = find_and_map method_match method_add_dist method_list
  (* in get_first_candidate candidates *)
  in match candidates with
    | [] -> raise (Type_error(UndeclaredMethod(name)))
    | [ (d,m) ] -> m  (* return method *)
    | _ ->
        begin
          if verbose then
            begin
              Printf.printf "Resolving call to %s with candidates : \n%!" method_2_call;
              let disp m = Printf.printf "  %s\n" (get_mangled_name current_class (snd m))
              in List.iter disp candidates
            end ;
          try
            let algo = match algorithm with
              | First -> get_first_candidate
              | Java_1_2 | Java_1_5 -> get_add_candidate
              | Multiply -> get_mult_candidate
              | Perfect -> get_perfect_candidate
            in let res = algo current_class candidates
            in (if verbose then Printf.printf "    => %s\n%!" (get_mangled_name current_class res)) ;
              res
          (* TODO : custom typing error *)
          with
          | Not_found -> raise (Type_error(UndeclaredMethod(name)))
          | MultipleMethods(l) -> raise(Type_error(AmbiguousCall(get_algorithm_name algorithm, method_2_call, l)))
        end
;;



