%{
  open Ast;;

  let dispatch_decl l =
    let rec f c i = function
      | [] -> (c, i)
      | `class_decl cl :: tl -> f (cl :: c) i tl
      | `interf_decl it :: tl -> f c (it :: i) tl
    in
      f [] [] l

  let dispatch_class_body l =
    let rec g f m s c = function
      | [] -> (f, m, s, c)
      | `field_declaration fd :: tl -> g (fd @ f) m s c tl
      | `method_declaration md :: tl -> g f (md :: m) s c tl
      | `static_initializer si :: tl -> g f m (si :: s) c tl
      | `constructor_declaration cd :: tl -> g f m s (cd :: c) tl
    in
      g [] [] [] [] l

  let dispatch_interf_body l =
    let rec f c a = function
      | [] -> (c, a)
      | `constant_declaration cd :: tl -> f (cd @ c) a tl
      | `abstract_declaration ad :: tl -> f c (ad :: a) tl
    in
      f [] [] l

  let expr_of_varlist l =
    let rec eov = function
      | [] -> assert false
      | [ x ] -> Evar x
      | h::t -> Efield_access((eov t),  h)
    in eov (List.rev l)

  let expr_of_affect a =
    match a with
      | Aname l -> expr_of_varlist l
      | Afield (e, s) -> Efield_access (e, s)
      | Aarray (e1, e2) -> Earray_access (e1, e2)

  let rec array_type_merge t a =
    if a = 0 then t else array_type_merge (`array t) (pred a)

  let rec array_type_option_merge t a =
    match t with
      | Some t -> Some (array_type_merge t a)
      | None ->
	  if a <> 0 then
	    raise (Failure "void method declared with []")
	  else
	    None
%}

%token <char> CHARACTER
%token <int> INTEGER
%token <float> FLOATING
%token <string> IDENT DOC STRING
%token DOT COMMA COLON SEMICOLON
%token EQUAL PLUS MINUS MUL DIV MOD OR AND XOR TILD INF_INF SUP_SUP SUP_SUP_SUP
%token AND_AND OR_OR PLUS_EQUAL MINUS_EQUAL MUL_EQUAL DIV_EQUAL MOD_EQUAL
%token OR_EQUAL AND_EQUAL XOR_EQUAL TILD_EQUAL INF_INF_EQUAL SUP_SUP_EQUAL
%token SUP_SUP_SUP_EQUAL AND_AND_EQUAL OR_OR_EQUAL PLUS_PLUS MINUS_MINUS
%token EQUAL_EQUAL NOT_EQUAL SUP SUP_EQUAL INF INF_EQUAL NOT INTERG
%token LEFTPAR RIGHTPAR LEFTBRACKET RIGHTBRACKET LEFTBRACE RIGHTBRACE
%token PACKAGE IMPORT CLASS EXTENDS IMPLEMENTS INTERFACE RETURN THROW BREAK
%token CONTINUE IF ELSE DO WHILE FOR TRY CATCH FINALLY SWITCH CASE DEFAULT
%token NULL SUPER THIS INSTANCEOF TRUE FALSE NEW BOOLEAN BYTE CHAR SHORT INT
%token FLOAT LONG DOUBLE PUBLIC PRIVATE PROTECTED STATIC FINAL NATIVE
%token SYNCHRONIZED ABSTRACT THREADSAFE TRANSIENT VOLATILE THROWS VOID
%token EOF PRINTLN

%type <Ast.compil_unit> main
%start main

%type <Ast.expression> expression
%start expression

%type <Ast.instruction> statement
%start statement

%type <Ast.instruction list> block_statements_opt
%start block_statements_opt

%%

main: compilation_unit { $1 };

literal:
  | INTEGER   { Eint $1 }
  | FLOATING  { Efloat $1 }
  | CHARACTER { Echar (int_of_char $1) }
  | STRING    { Estring $1 }
  | TRUE      { Ebool true }
  | FALSE     { Ebool false }
  | NULL      { Enull }
;

type_:
  | primitive_type { $1 }
  | reference_type { $1 }
;

primitive_type:
  | numeric_type { $1 }
  | BOOLEAN      { `boolean }
;

numeric_type:
  | integral_type       { $1 }
  | floating_point_type { $1 }
;

integral_type:
  | BYTE  { `byte  }
  | SHORT { `short }
  | INT   { `int   }
  | LONG  { `long  }
  | CHAR  { `char  }
;

floating_point_type:
  | FLOAT  { `float }
  | DOUBLE { `double }
;

reference_type:
  | class_or_interface_type { `java_object $1 }
  | array_type              { `array $1 }
;

class_or_interface_type:
  | name { $1 }
;

class_type:
  | class_or_interface_type { $1 }
;

interface_type:
  | class_or_interface_type { $1 }
;

array_type:
  | primitive_type LEFTBRACKET RIGHTBRACKET { $1 }
  | name LEFTBRACKET RIGHTBRACKET           { `java_object $1 }
  | array_type LEFTBRACKET RIGHTBRACKET     { `array $1 }
;

name:
  | simple_name    { [ $1 ] }
  | qualified_name { $1 }
;

simple_name:
  | IDENT { $1 }
;

qualified_name:
  | name DOT IDENT { $1 @ [ $3 ] }
;

compilation_unit:
  | package_declaration_opt import_declarations type_declarations {
      let (c, i) = dispatch_decl $3 in
	{
	  package = $1;
	  imports = $2;
	  class_decls = c;
	  interf_decls = i;
	}
    }
;

import_declarations:
  |                                        { [] }
  | import_declarations import_declaration { $2 :: $1 }
;

type_declarations:
  |                                    { [] }
  | type_declarations type_declaration {
      match $2 with
	| Some d -> d :: $1
	| None -> $1
    }
;

package_declaration_opt:
  |                        { None }
  | PACKAGE name SEMICOLON { Some $2 }
;

import_declaration:
  | single_type_import_declaration    { $1 }
  | type_import_on_demand_declaration { $1 }
;

single_type_import_declaration:
  | IMPORT name SEMICOLON { Isingle $2 }
;

type_import_on_demand_declaration:
  | IMPORT name DOT MUL SEMICOLON { Iall $2 }
;

type_declaration:
  | class_declaration     { Some $1 }
  | interface_declaration { Some $1 }
  | SEMICOLON             { None }
;

modifiers:
  |                    { [] }
  | modifier modifiers { $1 :: $2 }
;

modifier:
  | PUBLIC       { `public       }
  | PROTECTED    { `protected    }
  | PRIVATE      { `private_     }
  | STATIC       { `static       }
  | ABSTRACT     { `abstract     }
  | FINAL        { `final        }
  | NATIVE       { `native       }
  | SYNCHRONIZED { `synchronized }
  | TRANSIENT    { `transient    }
  | VOLATILE     { `volatile     }
;

class_declaration:
  | modifiers CLASS IDENT extends interfaces class_body {
      let (f, m, s, c) = dispatch_class_body $6 in
	`class_decl {
	  class_modifiers = $1;
	  class_name = $3;
	  class_extend = $4;
	  class_implements = $5;
	  class_fields = f;
	  class_methods = m;
	  class_initializers = s;
	  class_constructors = c;
	}
    }
;

extends:
  |                    { [ "Object" ] }
  | EXTENDS class_type { $2 }
;

interfaces:
  |                                { [] }
  | IMPLEMENTS interface_type_list { $2 }
;

interface_type_list:
  | interface_type                           { [ $1 ] }
  | interface_type_list COMMA interface_type { $3 :: $1 }
;

class_body:
  | LEFTBRACE class_body_declarations RIGHTBRACE { $2 }
;

class_body_declarations:
  |                                                { [] }
  | class_body_declarations class_body_declaration { $2 :: $1 }
;

class_body_declaration:
  | class_member_declaration { $1 }
  | static_initializer       { `static_initializer $1 }
  | constructor_declaration  { `constructor_declaration $1 }
;

class_member_declaration:
  | field_declaration  { `field_declaration $1 }
  | method_declaration { `method_declaration $1 }
;

field_declaration:
  | modifiers type_ variable_declarators SEMICOLON {
      let f ((name, arity), init) =  {
	field_modifiers = $1;
	field_type = array_type_merge $2 arity;
	field_name = name;
	field_initializer = init;
      } in
	List.map f $3
    }
;

variable_declarators:
  | variable_declarator                            { [ $1 ] }
  | variable_declarators COMMA variable_declarator { $3 :: $1 }
;

variable_declarator:
  | variable_declarator_id                            { ($1, None) }
  | variable_declarator_id EQUAL variable_initializer { ($1, Some $3) }
;

variable_declarator_id:
  | IDENT                                           { ($1, 0) }
  | variable_declarator_id LEFTBRACKET RIGHTBRACKET { (fst $1, snd $1 + 1) }
;

variable_initializer:
  | expression        { Simple_init $1 }
  | array_initializer { Array_init $1 }
;

method_declaration:
  | method_header method_body {
      let (m, r, ((i, pl), a), t) = $1 in
	{
	  method_modifiers = m;
	  method_return = array_type_option_merge r a;
	  method_name = i;
	  method_parameters = List.rev pl;
	  method_throws = t;
	  method_body = $2;
	}
    }
;

method_header:
  | modifiers type_ method_declarator throws { ($1, Some $2, $3, $4) }
  | modifiers VOID method_declarator throws  { ($1, None, $3, $4) }
;

method_declarator:
  | IDENT LEFTPAR formal_parameter_list_opt RIGHTPAR { (($1, $3), 0) }
  | method_declarator LEFTBRACKET RIGHTBRACKET       { (fst $1, snd $1 + 1) }
;

formal_parameter_list_opt: { [] } | formal_parameter_list { $1 }
formal_parameter_list:
  | formal_parameter                             { [ $1 ] }
  | formal_parameter_list COMMA formal_parameter { $3 :: $1 }
;

formal_parameter:
  | type_ variable_declarator_id { (array_type_merge $1 (snd $2), fst $2) }
;

throws:
  |                        { [] }
  | THROWS class_type_list { $2 }
;

class_type_list:
  | class_type                       { [ $1 ] }
  | class_type_list COMMA class_type { $3 :: $1 }
;

method_body:
  | block     { Some $1 }
  | SEMICOLON { None }
;

static_initializer:
  | STATIC block { $2 }
;

constructor_declaration:
  | modifiers constructor_declarator throws constructor_body {
      {
	constructor_modifiers = $1;
	constructor_name = fst $2;
	constructor_parameters = List.rev (snd $2);
	constructor_throws = $3;
	constructor_call = fst $4;
	constructor_body = snd $4;
      }
    }
;

constructor_declarator:
  | simple_name LEFTPAR formal_parameter_list_opt RIGHTPAR { ($1, $3) }
;

constructor_body:
  | LEFTBRACE RIGHTBRACE                                 { (Cnone, Some []) }
  | LEFTBRACE block_statements RIGHTBRACE                { (Cnone, Some $2) }
  | LEFTBRACE explicit_constructor_invocation RIGHTBRACE { ($2, None) }
  | LEFTBRACE explicit_constructor_invocation block_statements RIGHTBRACE {
      ($2, Some $3)
    }
;

explicit_constructor_invocation:
  | THIS LEFTPAR argument_list RIGHTPAR SEMICOLON  { Cthis $3 }
  | SUPER LEFTPAR argument_list RIGHTPAR SEMICOLON { Csuper $3 }
;

interface_declaration:
  | modifiers INTERFACE IDENT extends_interfaces_opt interface_body {
      let (c, a) = dispatch_interf_body $5 in
	`interf_decl {
	  interf_modifiers = $1;
	  interf_name = $3;
	  interf_extends = $4;
	  interf_constants = c;
	  interf_abstracts = a;
	}
    }
;

extends_interfaces_opt: { [] } | extends_interfaces { $1 };
extends_interfaces:
  | EXTENDS interface_type                  { [ $2 ] }
  | extends_interfaces COMMA interface_type { $3 :: $1 }
;

interface_body:
  | LEFTBRACE interface_member_declarations RIGHTBRACE { $2 }
;

interface_member_declarations:
  |                                                            { [] }
  | interface_member_declarations interface_member_declaration { $2 :: $1 }
;

interface_member_declaration:
  | constant_declaration        { `constant_declaration $1 }
  | abstract_method_declaration { `abstract_declaration $1 }
;

constant_declaration:
  | field_declaration { $1 }
;

abstract_method_declaration:
  | method_header SEMICOLON {
      let (m, r, ((n, pl), a), t) = $1 in
	{
	  abstract_modifiers = m;
	  abstract_return = array_type_option_merge r a;
	  abstract_name = n;
	  abstract_parameters = List.rev pl;
	  abstract_throws = t;
	}
    }
;

array_initializer:
  | LEFTBRACE variable_initializers comma_opt RIGHTBRACE { $2 }
;

comma_opt: { () } | COMMA { () };

variable_initializers:
  |                                                  { [] }
  | variable_initializer COMMA variable_initializers { $1 :: $3 }
;

block:
  | LEFTBRACE block_statements_opt RIGHTBRACE { $2 }
;

block_statements_opt: { [] } | block_statements { $1 };
block_statements:
  | block_statement                  { $1 }
  | block_statements block_statement { $1 @ $2 }
;

block_statement:
  | local_variable_declaration_statement { $1 }
  | statement                            { [ $1 ] }
;

local_variable_declaration_statement:
  | local_variable_declaration SEMICOLON {
      let f (t, n, i) = Idecl (t, n, i) in
	List.map f $1
    }
;

local_variable_declaration:
  | type_ variable_declarators {
      let f ((n, a), i) = (array_type_merge $1 a, n, i) in
	List.rev_map f $2
    }
;

statement:
  | statement_without_trailing_substatement { $1 }
  | labeled_statement                       { $1 }
  | if_then_statement                       { $1 }
  | if_then_else_statement                  { $1 }
  | while_statement                         { $1 }
  | for_statement                           { $1 }
;

statement_no_short_if:
  | statement_without_trailing_substatement { $1 }
  | labeled_statement_no_short_if           { $1 }
  | if_then_else_statement_no_short_if      { $1 }
  | while_statement_no_short_if             { $1 }
  | for_statement_no_short_if               { $1 }
;

statement_without_trailing_substatement:
  | block                  { Iblock $1 }
  | empty_statement        { $1 }
  | expression_statement   { $1 }
  | switch_statement       { $1 }
  | do_statement           { $1 }
  | break_statement        { $1 }
  | continue_statement     { $1 }
  | return_statement       { $1 }
  | synchronized_statement { $1 }
  | throw_statement        { $1 }
  | try_statement          { $1 }
  | println_statement      { $1 }
;

println_statement:
  | PRINTLN LEFTPAR expression RIGHTPAR SEMICOLON { Iprintln $3 }
;

empty_statement:
  | SEMICOLON { Inop }
;

labeled_statement:
  | IDENT COLON statement { Ilabeled ($1, $3) }
;

labeled_statement_no_short_if:
  | IDENT COLON statement_no_short_if { Ilabeled ($1, $3) }
;

expression_statement:
  | statement_expression SEMICOLON { Iexpr $1 }
;

statement_expression:
  | assignment                         { $1 }
  | pre_increment_expression           { $1 }
  | pre_decrement_expression           { $1 }
  | post_increment_expression          { $1 }
  | post_decrement_expression          { $1 }
  | method_invocation                  { $1 }
  | class_instance_creation_expression { $1 }
;

if_then_statement:
  | IF LEFTPAR expression RIGHTPAR statement { Iifthenelse ($3, $5, Inop) }
;

if_then_else_statement:
  | IF LEFTPAR expression RIGHTPAR statement_no_short_if ELSE statement {
      Iifthenelse ($3, $5, $7)
    }
;

if_then_else_statement_no_short_if:
  | IF LEFTPAR expression RIGHTPAR statement_no_short_if ELSE
      statement_no_short_if {
	Iifthenelse ($3, $5, $7)
      }
;

switch_statement:
  | SWITCH LEFTPAR expression RIGHTPAR switch_block {
      Iswitch ($3, $5)
    }
;

switch_block:
  | LEFTBRACE switch_labels RIGHTBRACE {
      [ ($2, [ Inop ]) ]
    }
  | LEFTBRACE switch_block_statement_groups RIGHTBRACE {
      $2
    }
  | LEFTBRACE switch_block_statement_groups switch_labels RIGHTBRACE {
      ($3, [ Inop ]) :: $2
    }
;

switch_block_statement_groups:
  | switch_block_statement_group                               { [ $1 ] }
  | switch_block_statement_groups switch_block_statement_group { $2 :: $1 }
;

switch_block_statement_group:
  | switch_labels block_statements { ($1, $2) }
;

switch_labels:
  | switch_label               { [ $1 ] }
  | switch_labels switch_label { $2 :: $1 }
;

switch_label:
  |  CASE constant_expression COLON { Scase $2 }
  |  DEFAULT COLON                  { Sdefault }
;

while_statement:
  | WHILE LEFTPAR expression RIGHTPAR statement { Iwhile ($3, $5) }
;

while_statement_no_short_if:
  | WHILE LEFTPAR expression RIGHTPAR statement_no_short_if {
      Iwhile ($3, $5)
    }
;

do_statement:
  | DO statement WHILE LEFTPAR expression RIGHTPAR SEMICOLON {
      Idowhile ($2, $5)
    }
;

for_statement:
  | FOR LEFTPAR for_init_opt SEMICOLON expression_opt SEMICOLON
      for_update_opt RIGHTPAR statement {
	Ifor ($3, $5, $7, $9)
      }
;

for_statement_no_short_if:
  | FOR LEFTPAR for_init_opt SEMICOLON expression_opt SEMICOLON
      for_update_opt RIGHTPAR statement_no_short_if {
	Ifor ($3, $5, $7, $9)
      }
;

for_init_opt: { Fnone } | for_init { $1 };
for_init:
  | statement_expression_list  { Fexprs $1 }
  | local_variable_declaration { Fdecls $1 }
;

for_update_opt: { [] } | for_update { $1 };
for_update:
  | statement_expression_list { $1 }
;

statement_expression_list:
  | statement_expression                                 { [ $1 ] }
  | statement_expression COMMA statement_expression_list { $1 :: $3 }
;

break_statement:
  | BREAK identifier_opt SEMICOLON { Ibreak $2 }
;

identifier_opt: { None } | IDENT { Some $1 };

continue_statement:
  | CONTINUE identifier_opt SEMICOLON { Icontinue $2 }
;

return_statement:
  | RETURN expression_opt SEMICOLON { Ireturn $2 }
;

throw_statement:
  | THROW expression SEMICOLON { Ithrow $2 }
;

synchronized_statement:
  | SYNCHRONIZED LEFTPAR expression RIGHTPAR block { Isynchronized ($3, $5) }
;

try_statement:
  | TRY block catches {
      let f ((t, n), b) = (t, n, b) in
	Itry ($2, List.map f $3, None)
    }
  | TRY block catches_opt finally {
      let f ((t, n), b) = (t, n, b) in
	Itry ($2, List.map f $3, Some $4)
    }
;

catches_opt: { [] } | catches { $1 };
catches:
  | catch_clause         { [ $1 ] }
  | catches catch_clause { $2 :: $1 }
;

catch_clause:
  | CATCH LEFTPAR formal_parameter RIGHTPAR block { ($3, $5) }
;

finally:
  | FINALLY block { $2 }
;

primary:
  | primary_no_new_array      { $1 }
  | array_creation_expression { $1 }
;

primary_no_new_array:
  | literal                            { $1 }
  | THIS                               { Ethis }
  | LEFTPAR expression RIGHTPAR        { $2 }
  | class_instance_creation_expression { $1 }
  | field_access                       { Efield_access (fst $1, snd $1) }
  | method_invocation                  { $1 }
  | array_access                       { Earray_access (fst $1, snd $1) }
;

class_instance_creation_expression:
  | NEW class_type LEFTPAR argument_list RIGHTPAR { Enew_object ($2, $4) }
;

argument_list:
  |                                { [] }
  | expression                     { [ $1 ] }
  | expression COMMA argument_list { $1 :: $3 }
;

array_creation_expression:
  | NEW primitive_type dim_exprs dims_opt          { Enew_array ($2, $3, $4) }
  | NEW class_or_interface_type dim_exprs dims_opt {
      Enew_array (`java_object $2, $3, $4)
    }
;

dim_exprs:
  | dim_expr           { [ $1 ] }
  | dim_exprs dim_expr { $2 :: $1 }
;

dim_expr:
  | LEFTBRACKET expression RIGHTBRACKET { $2 }
;

dims_opt: { 0 } | dims { $1 };
dims:
  | LEFTBRACKET RIGHTBRACKET      { 1 }
  | dims LEFTBRACKET RIGHTBRACKET { succ $1 }
;

field_access:
  | primary DOT IDENT { ($1, $3) }
  | SUPER DOT IDENT   { (Esuper, $3) }
;

method_invocation:
  | name LEFTPAR argument_list RIGHTPAR {
      let m = List.hd $1 in
      let o = List.tl $1 in
	if o = [] then Ecall (m, $3) else
    (
      let rec get_last l =
        match l with
          | [] -> failwith "list should not be empty..."
          | h::[] -> h
          | h::t -> get_last t
      and rem_last l = 
        match l with
          | [] -> failwith "list should not be empty..."
          | h::[] -> []
          | h::t -> h::(rem_last t)
          
    in Eexpr_call (expr_of_varlist (rem_last $1), (get_last $1), $3)
    )
    }
  | primary DOT IDENT LEFTPAR argument_list RIGHTPAR {
      Eexpr_call ($1, $3, $5)
    }
  | SUPER DOT IDENT LEFTPAR argument_list RIGHTPAR {
      Eexpr_call (Esuper, $3, $5)
    }
;

array_access:
  | name LEFTBRACKET expression RIGHTBRACKET {
      let t = List.hd $1 in
      let o = List.tl $1 in
	if o = [] then (Evar t, $3) else
	  (expr_of_varlist $1, $3)
    }
  | primary_no_new_array LEFTBRACKET expression RIGHTBRACKET {
      ($1, $3)
    }
;

postfix_expression:
  | primary                   { $1 }
  | name                      { expr_of_varlist $1 }
  | post_increment_expression { $1 }
  | post_decrement_expression { $1 }
;

post_increment_expression:
  | postfix_expression PLUS_PLUS { Eunary (Opostinc, $1) }
;

post_decrement_expression:
  | postfix_expression MINUS_MINUS { Eunary (Opostdec, $1) }
;

unary_expression:
  | pre_increment_expression        { $1 }
  | pre_decrement_expression        { $1 }
  | PLUS unary_expression           { $2 }
  | MINUS unary_expression          { Eunary (Oneg, $2) }
  | unary_expression_not_plus_minus { $1 }
;

pre_increment_expression:
  | PLUS_PLUS unary_expression { Eunary (Opreinc, $2) }
;

pre_decrement_expression:
  | MINUS_MINUS unary_expression { Eunary (Opredec, $2) }
;

unary_expression_not_plus_minus:
  | postfix_expression    { $1 }
  | TILD unary_expression { Eunary (Otild, $2) }
  | NOT unary_expression  { Eunary (Onot, $2) }
  | cast_expression       { $1 }
;

cast_expression:
  | LEFTPAR primitive_type dims_opt RIGHTPAR unary_expression {
      Etype_cast (array_type_merge $2 $3, $5)
    }
  | LEFTPAR expression RIGHTPAR unary_expression_not_plus_minus {
      Eexpr_cast ($2, $4)
    }
  | LEFTPAR name dims RIGHTPAR unary_expression_not_plus_minus {
      Etype_cast (array_type_merge (`java_object $2) $3, $5)
    }
;

multiplicative_expression:
  | unary_expression                               { $1 }
  | multiplicative_expression MUL unary_expression { Ebinary (Omul, $1, $3) }
  | multiplicative_expression DIV unary_expression { Ebinary (Odiv, $1, $3) }
  | multiplicative_expression MOD unary_expression { Ebinary (Omod, $1, $3) }
;

additive_expression:
  | multiplicative_expression { $1 }
  | additive_expression PLUS multiplicative_expression {
      Ebinary (Oplus, $1, $3)
    }
  | additive_expression MINUS multiplicative_expression {
      Ebinary (Ominus, $1, $3)
    }
;

shift_expression:
  | additive_expression { $1 }
  | shift_expression INF_INF additive_expression {
      Ebinary (Oinfinf, $1, $3)
    }
  | shift_expression SUP_SUP additive_expression {
      Ebinary (Osupsup, $1, $3)
    }
  | shift_expression SUP_SUP_SUP additive_expression {
      Ebinary (Osupsupsup, $1, $3)
    }
;

relational_expression:
  | shift_expression { $1 }
  | relational_expression INF shift_expression {
      Ebincomp (Oinf, $1, $3)
    }
  | relational_expression SUP shift_expression {
      Ebincomp (Osup, $1, $3)
    }
  | relational_expression INF_EQUAL shift_expression {
      Ebincomp (Oinfequal, $1, $3)
    }
  | relational_expression SUP_EQUAL shift_expression {
      Ebincomp (Osupequal, $1, $3)
    }
  | relational_expression INSTANCEOF reference_type {
      Einstanceof ($1, $3)
    }
;

equality_expression:
  | relational_expression { $1 }
  | equality_expression EQUAL_EQUAL relational_expression {
      Ebincomp (Oequal, $1, $3)
    }
  | equality_expression NOT_EQUAL relational_expression {
      Ebincomp (Onotequal, $1, $3)
    }
;

and_expression:
  | equality_expression                    { $1 }
  | and_expression AND equality_expression { Ebinary (Oand, $1, $3) }
;

exclusive_or_expression:
  | and_expression                             { $1 }
  | exclusive_or_expression XOR and_expression { Ebinary (Oxor, $1, $3) }
;

inclusive_or_expression:
  | exclusive_or_expression                            { $1 }
  | inclusive_or_expression OR exclusive_or_expression { Ebinary (Oor, $1, $3) }
;

conditional_and_expression:
  | inclusive_or_expression { $1 }
  | conditional_and_expression AND_AND inclusive_or_expression {
      Ebinary (Oandand, $1, $3)
    }
;

conditional_or_expression:
  | conditional_and_expression { $1 }
  | conditional_or_expression OR_OR conditional_and_expression {
      Ebinary (Ooror, $1, $3)
    }
;

conditional_expression:
  | conditional_or_expression { $1 }
  | conditional_or_expression INTERG expression COLON conditional_expression {
      Eternary (Oifthenelse, $1, $3, $5)
    }
;

assignment_expression:
  | conditional_expression { $1 }
  | assignment             { $1 }
;

assignment:
  | left_hand_side assignment_operator assignment_expression {
      match $2 with
	| `equal -> Eaffect ($1, $3)
	| `mul_equal -> Eaffect ($1, Ebinary (Omul, expr_of_affect $1, $3))
	| `div_equal -> Eaffect ($1, Ebinary (Odiv, expr_of_affect $1, $3))
	| `mod_equal -> Eaffect ($1, Ebinary (Omod, expr_of_affect $1, $3))
	| `plus_equal -> Eaffect ($1, Ebinary (Oplus, expr_of_affect $1, $3))
	| `minus_equal -> Eaffect ($1, Ebinary (Ominus, expr_of_affect $1, $3))
	| `inf_inf_equal ->
	    Eaffect ($1, Ebinary (Oinfinf, expr_of_affect $1, $3))
	| `sup_sup_equal ->
	    Eaffect ($1, Ebinary (Osupsup, expr_of_affect $1, $3))
	| `sup_sup_sup_equal ->
	    Eaffect ($1, Ebinary (Osupsupsup, expr_of_affect $1, $3))
	| `and_equal -> Eaffect ($1, Ebinary (Oand, expr_of_affect $1, $3))
	| `xor_equal -> Eaffect ($1, Ebinary (Oxor, expr_of_affect $1, $3))
	| `or_equal -> Eaffect ($1, Ebinary (Oor, expr_of_affect $1, $3))
    }
;

left_hand_side:
  | name         { Aname $1 }
  | field_access { Afield (fst $1, snd $1) }
  | array_access { Aarray (fst $1, snd $1) }
;

assignment_operator:
  | EQUAL             { `equal }
  | MUL_EQUAL         { `mul_equal }
  | DIV_EQUAL         { `div_equal }
  | MOD_EQUAL         { `mod_equal }
  | PLUS_EQUAL        { `plus_equal }
  | MINUS_EQUAL       { `minus_equal }
  | INF_INF_EQUAL     { `inf_inf_equal }
  | SUP_SUP_EQUAL     { `sup_sup_equal }
  | SUP_SUP_SUP_EQUAL { `sup_sup_sup_equal }
  | AND_EQUAL         { `and_equal }
  | XOR_EQUAL         { `xor_equal }
  | OR_EQUAL          { `or_equal }
;

expression_opt: { None } | expression { Some $1 };
expression:
  | assignment_expression { $1 }
;

constant_expression:
  | expression { $1 }
;
