
open Ast;;
open Ast_typer;;
open Overloader;;

type typer_options = {
  mutable verbose : bool ;
  mutable overloader : Overloader.overloading_algorithm
};;

let type_options = {
  verbose = false;
  overloader = Overloader.Java_1_5;
};;

let blank_typed_method =
  {
    typed_method_modifiers = [];
    typed_method_return = None;
    typed_method_name = "";
    typed_method_mangled_name = "";
    typed_method_parameters = [];
    typed_method_throws = [];
    typed_method_body = None;
  };;

let create_default_constructor name call_super =
  {
    constructor_modifiers = [];
    constructor_name = name;
    constructor_parameters = [];
    constructor_throws = [];
    constructor_call = if call_super then Ast.Csuper([]) else Ast.Cnone;
    constructor_body = None;
  }

let obj_class = {
  class_modifiers = [];
  class_name = "Object";
  class_extend = [];
  class_implements = [];
  class_fields = [];
  class_methods = [];
  class_initializers = [];
  class_constructors = [];
};;

let obj = {
  ic_name = "Object";
  ic_abstract = true;
  ic_father = None;
  ic_children = [];
  ic_class = obj_class;
  ic_fields = [];
  ic_methods = [];
  ic_cons = [ (0, create_default_constructor "Object" false) ];
};;

(* ENVIRONMENT *)
type env = {
  env_current_class : inter_class;
  mutable env_current_method : method_def;
  env_field_decls : typed_field list; (* typed fields *)
  env_method_def : method_def list; (* method list... *)
  env_hash_var : (string, typed_affect) Hashtbl.t;       (* current variable stack *)
  env_hash_meth : (string, env_method_def) Hashtbl.t;
  env_block_depth : int ref;        (* current block depth *)
  env_block_var_id : ((int ref) list) ref; (* stack of (next local var id in block)  *)
  env_classes : inter_class;
  (* env_overloader : overloading_algorithm; *)
}

and env_method_def = {
  env_method_modifiers : modifier list;
  env_method_return : java_type option;
  env_method_parameters : (java_type * string) list;
  env_method_throws : class_type list;
};;

let empty_method = {
  method_modifiers = [];
  method_return = None;
  method_name = "";
  method_parameters = [];
  method_throws = [];
  method_body = None;
};;

let empty_env = {
  env_current_class = obj;
  env_current_method = empty_method;
  env_field_decls = [] ;
  env_method_def = [] ;
  env_hash_var = Hashtbl.create 16;
  env_hash_meth = Hashtbl.create 16;
  env_block_depth = ref 0;
  env_block_var_id = ref [ref 0];
  env_classes = obj;
};;

(* let hd l = try List.hd l with | Failure _ -> assert *)

let get_type_default_value t =
  match t with
  | `boolean -> `boolean, Ebool(false)
  | `byte | `char | `short | `int -> t, Eint(0)
  | `float -> `float, Efloat(0.0)
  | `string -> `string, Estring("")
  | `java_object(ct) -> `java_object(ct), Enull
  | `array(jt) -> `array(jt), Enull
  | _ -> raise (Type_error(UnexpectedType(t)));; (* should not be here *)

let get_class_parent c =
  match c.ic_father with
  | None -> failwith "No parent !"
  | Some(p) -> p;;


(* CHECK TYPE VALIDITY OF OPERATOR ARGUMENTS a type is valid if it can     *)
(* match with one of the types in a given list                             *)
let is_type_valid t l =
  let is_type_compatible t1 t2 =
    match t1 with
    | t when t1 = t2 -> true
    | `byte | `char | `short | `int -> (t2 = `int)
    | `float | `double -> (t2 = `float)
(*    | `string -> (t2 = `string)
    | `boolean -> (t2 = `boolean) *)
    | `java_object(_) -> (match t2 with | `java_object(_) -> true | _ -> false)
    | `array(_) -> (t2 = `array(`int))
    | `nil -> (t2 = `java_object([]))
    | _ -> false
  in if (List.exists (is_type_compatible t) l)
    then ()
    else raise (Type_error(IncompatibleTypes(t, l)));;

(* binary comparator is valid if type t is valid. returns out_type *)
let check_unary_op op t =
  let l = match op with
    | Oneg -> [ `int ; `float ]
    | Onot -> [ `int ; `boolean ]
    | Otild -> [ `int ]
    | Opreinc | Opredec | Opostinc | Opostdec -> [ `int ] (* changer, c'est faux *)
  in is_type_valid t l ; t;;

let check_binary_comp op t =
  (match op with
    | Oequal | Onotequal -> is_type_valid t [ `boolean ; `int ; `float ; `string ; `java_object([]) ; `array(`int) ]
    | _ -> is_type_valid t [ `int ; `float ]
  ) ; `boolean;;

let check_binary_op op t =
  let l = match op with
    | Oplus -> [ `int ; `float ; `string ]
    | Ominus | Omul | Odiv -> [ `int ; `float ]
    | Omod | Oinfinf | Osupsup | Osupsupsup -> [ `int ]
    | Oor | Oand | Oxor -> [ `int ; `boolean ]
    | Oandand | Ooror -> [ `boolean ]
  in is_type_valid t l ; t;;

(* TYPE MANAGEMENT *)
let rec restrict_type t =
  match t with
  | `long -> `int
  | `double -> `float
  | `java_object(o) when (String.compare (List.hd o) "String") = 0 -> `string
  | `array(jt) -> `array(restrict_type jt)
  | _ -> t;;

let get_method env name nb_param =
  (* TODO : nb param... check in env hashtbl... *)
  let match_method_name meth = ((String.compare name meth.method_name) = 0)
  in
    try
      List.find match_method_name env.env_method_def
    with | Not_found -> raise (Type_error(UndeclaredMethod(name)));;

let mangle_method name params =
  let get_type m = fst m
  in let rec name_param l =
    match l with
    | [] -> ""
    | h::[] -> string_of_java_type (get_type h) ^ " "
    | h:: t -> (string_of_java_type (get_type h)) ^ ", " ^ name_param t
  in
    name ^ "( " ^ (name_param params) ^ ")";;

let method_2_typelist (class_id, method_id, mangled_name, met) =
  let get_p p = restrict_type (fst p)
  in let get_parameter_types p = List.map get_p p
  in let param_type_list = (get_parameter_types met.method_parameters)
  in (class_id, method_id, met.method_name, mangled_name, param_type_list)
;;

let cons_2_typelist (cons_id, cons) =
  let get_p p = restrict_type (fst p)
  in let get_parameter_types p = List.map get_p p
  in let param_type_list = (get_parameter_types cons.constructor_parameters)
  in let mangled_name = (mangle_method cons.constructor_name cons.constructor_parameters)
  in (0, cons_id, cons.constructor_name, mangled_name, param_type_list)
;;

(* FUNCTION OVERLOADING (surcharge...) *)
let get_class_method env c method_name param_types =
  let methods_list = List.map method_2_typelist c.ic_methods
  in let (_, method_id, _, _, _) =
    overload type_options.verbose type_options.overloader c env.env_classes method_name param_types methods_list
  in let method_has_id id (_, i, _, _) = (id = i)
  in List.find (method_has_id method_id) c.ic_methods;;

let get_class_method_from_id c id =
  let valid_m (_, i, _, m) = (i = id)
  in let (_, _, _, m) = List.find valid_m c.ic_methods
  in m;;

let get_class_field c name =
  let is_field f =
    (String.compare f.typed_field_name name) = 0
  in try
      let f = List.find is_field c.ic_fields
      in Afield(f.typed_field_type, f.typed_field_super_id, f.typed_field_id)
    with | Not_found -> raise (Type_error(UndeclaredVariable(name)));;

let rec get_field_in_var_list l =
  match l with
  | [] -> raise Not_found
  | h:: t ->
      begin
        match h with
        | Afield _ -> h
        | _ -> get_field_in_var_list t
      end;;

let is_method_xxx m xxx =
  List.exists (fun x -> x = xxx) m.method_modifiers;;

let is_method_static m = is_method_xxx m `static;;
let is_method_abstract m = is_method_xxx m `abstract;;

(* CHECK FOR IMPLICIT CASTS *)
let rec implicit_cast env e t2 =
  let t1 = fst e in
    match t1, t2 with
    | _, _ when t1 = t2 -> e
    | `byte, `int | `char, `int | `short, `int -> `int, Ecast(Int_of_int, e)
    | `byte, `short | `char, `short -> `short, Ecast(Int_of_int, e)
    | `byte, `char | `short, `char -> `char, Ecast(Int_of_int, e)
    | `byte, `float | `char, `float
    | `short, `float | `int, `float -> `float, Ecast(Float_of_int, e)
    | `byte, `string | `char, `string
    | `short, `string | `int, `string -> `string, Ecast(String_of_int, e)
    | `float, `string -> `string, Ecast(String_of_float, e)
    | `boolean, `string -> `string, Ecast(String_of_bool, e)
    | `java_object(_), `string -> type_expcall env e "toString" []
    (* ( Printf.printf "String_of %s\n%!" (string_of_java_type t1) ;       *)
    (* `string, Ecast(String_of_object, e) )                               *)
    | `java_object(o1), `java_object(o2)
    -> Printf.printf "obj vs obj\n" ; if is_instance_of env.env_classes (List.hd o1) (List.hd o2) then e
        else raise (Type_error(InvalidImplicitCast(t1, t2)))
    | `nil, `string -> `string, Estring("null")
    | `nil, `java_object(o) -> `java_object(o), Enull
    | _ -> raise (Type_error(InvalidImplicitCast(t1, t2)))

and implicit_cast_list env el tl =
  match el, tl with
  | [], [] -> []
  | (eh:: et), (th:: tt) -> (implicit_cast env eh th):: (implicit_cast_list env et tt)
  | _ -> raise (Failure("Parsing error, incompatible parameter number..."))

(* TYPE EXPRESSIONS *)
and type_expcall env te met params =
  let class_name = match (fst te) with
    | `java_object(o) -> List.hd o
    | t -> raise (Type_error(NotAnObject(t)))
  in let c = get_class class_name env.env_classes (* find the class... *)
  in
  (* Printf.printf " " ; Printer.print_expression stdout                   *)
  (* (Ast.Eexpr_call(exp, met, params)) ; Printf.printf "\n%!" ;           *)
  let return_type, mangled_name, method_id, t_params = type_gen_call env c met params
  in return_type, Eexpr_call(te, mangled_name, method_id, t_params)

and type_gen_call env c met params =
  let t_params = List.map (type_exp env) params
  in let pt p = restrict_type (fst p) in let param_types = List.map pt t_params (* get all types *)
  in let valid_method = get_class_method env c met param_types
  in match valid_method with
    | (class_id, method_id, mangled_name, method_def) ->
        begin
          let return_type = match method_def.method_return with
            | None -> `void
            | Some(t) -> restrict_type t
          in
          (* Printf.printf " => %s\n\n%!" mangled_name ; *)
          let typed_params = implicit_cast_list env t_params param_types
          in (return_type, mangled_name, method_id, typed_params)
        end

and type_exp env e =
  let rec type_e e =
    let rec type_2_exp e1 e2 =
      (* get types for the 2 expressions *)
      let te1 = type_e e1 and te2 = type_e e2
      in let t1 = fst te1 and t2 = fst te2
      in match (t1, t2) with  (* requires compatible types... *)
        | (_, _) when t1 = t2 -> (te1, te2)    (* identical types, return the typed expressions *)
        | (`string, _ ) -> ((te1), (implicit_cast env te2 `string ))
        | (_ , `string) -> ((implicit_cast env te1 `string), (te2))
        | (_, `float) -> ((implicit_cast env te1 `float), (te2))
        | (`float, _) -> ((te1), (implicit_cast env te2 `float))
        | (_, `int) -> ((implicit_cast env te1 `int), (te2))
        | (`int, _) -> ((te1), (implicit_cast env te2 `int))
        | (_, `short) -> ((implicit_cast env te1 `short), (te2))
        | (`short, _) -> ((te1), (implicit_cast env te2 `short))
        | (_, `char) -> ((implicit_cast env te1 `char), (te2))
        | (`char, _) -> ((te1), (implicit_cast env te2 `char))
        | (`nil, `java_object(_)) -> ((implicit_cast env te1 t2), te2)
        | (`java_object(_), `nil) -> (te1, (implicit_cast env te2 t1))
        | _ -> raise (Type_error(Clash(t1, t2)))
    
    and get_field te name =
      match snd te with
      | Ethis -> (* get string field from current class... *)
          begin
            try
              get_field_in_var_list (Hashtbl.find_all env.env_hash_var name)
            with
            | Not_found -> raise (Type_error(UndeclaredVariable(name)))
          end
      | _ ->
          begin
            let ct = match fst te with
              | `java_object(ct) -> ct
              | t -> raise (Type_error(NotAnObject(t)))
            in let c = get_class (List.hd ct) env.env_classes
            in get_class_field c name
          end
    
    and get_var_type var =
      match var with
      | Alocal(jt, _, _) | Afield(jt, _, _) | Aparam(jt, _) -> jt
      | _ -> failwith "Unmanaged affectation..."
    
    (* type affect : get variable ID and check types... *)
    and type_affect aff e =
      let var = match aff with
        | Ast.Aname (sl) ->
            if (List.length sl > 1) then failwith "Unmanaged affectation list..."
            else begin
              let name = List.hd sl
              in
              (* check type validity... *)
                begin
                  try
                    let v = (Hashtbl.find env.env_hash_var name) in match v with
                      | Alocal(jt, block_id, var_id)
                      -> Alocal(jt, !(env.env_block_depth) - block_id, var_id)
                      | Afield(jt, _, _) | Aparam(jt, _) -> v
                      | _ -> failwith "Unmanaged affectation..."
                  with | Not_found -> raise (Type_error(UndeclaredVariable(name)))
                end
              
            end
        | Ast.Afield(exp, name) -> get_field (type_e exp) name
        | Ast.Aarray(exp1, exp2) -> failwith "Unmanaged affectation Aarray...\n"
      in let t = get_var_type var
      in let exp = implicit_cast env e t in t, Eaffect(var, exp)
    
    and type_var str =
      try
        let var = Hashtbl.find env.env_hash_var str in
          match var with
          | Alocal(jt, block_id, var_id)
          -> jt, Evar(Alocal(jt, !(env.env_block_depth) - block_id, var_id))
          | Afield(jt, _, _) | Aparam(jt, _) -> jt, Evar(var)
          | _ -> raise (Failure "Unmanaged affectation...\n")
      with | Not_found -> raise (Type_error(UndeclaredVariable(str)))
    
    and type_field_access exp name =
      let te = type_e exp
      in let var = get_field te name
      in match var with
        | Afield(jt, _, id) -> jt, Efield_access(te, name, id)
        | _ -> assert false
    
    (* TYPE CALLS *)
    and type_call met el =
      (* Check if calling non-static from static method... *)
      let return_type, mangled_name, method_id, t_params = type_gen_call env env.env_current_class met el
      in (* Check if calling non-static from static method...*)
      let new_method = get_class_method_from_id env.env_current_class method_id
      in begin
          if (is_method_static env.env_current_method) && not (is_method_static new_method)
          then raise (Type_error(CallingNonStaticFromStatic(mangled_name)))
        end ;
        return_type, Ecall(mangled_name, method_id, t_params)
    
    and type_exp_call exp met params =
      let te = type_e exp
      in type_expcall env te met params
    
    and type_new ct params =
      let t = `java_object(ct)
      and c = get_class (List.hd ct) env.env_classes
      in (if c.ic_abstract then raise (Type_error(CannotInstantiateAbstractClass(c.ic_name))));
        let cons_id, typed_params = get_class_constructor env c params
        in t, Enew_object(ct, cons_id, typed_params)
    
    and type_super () =
      let c = env.env_current_class
      in let p = get_class_parent c
      in (`java_object([p.ic_name]), Esuper)
    
    and type_this () =
      let c = env.env_current_class
      in (`java_object([c.ic_name]), Ethis)
    
    and type_cast jt e =
      (* TODO : check valididty... *)
      let _, te = type_e e in jt, te
    
    and type_exp_cast name e =
      type_cast (`java_object([name])) e
    
    in
      begin
        match e with
        (* convert Ast types to TypedAst types *)
        | Ast.Ebool b -> `boolean, Ebool(b)
        | Ast.Ebyte i -> `byte, Eint(i)
        | Ast.Eshort i -> `short, Eint(i)
        | Ast.Echar i -> `char, Eint(i)
        | Ast.Eint i -> `int, Eint(i)
        | Ast.Efloat f | Ast.Edouble f -> `float, Efloat(f)
        | Ast.Estring s -> `string, Estring(s)
        
        (* convert operators... *)
        | Ast.Eunary ( op, e1) -> let te1 = type_e e1
            in check_unary_op op (fst te1), Eunary( op, te1 )
        
        | Ast.Ebinary (op, e1, e2) -> let e1e2 = type_2_exp e1 e2
            in let te1 = fst e1e2 and te2 = snd e1e2
            in check_binary_op op (fst te1) , Ebinary(op, te1, te2)
        
        | Ast.Ebincomp (op, e1, e2) -> let e1e2 = type_2_exp e1 e2
            in let te1 = fst e1e2 and te2 = snd e1e2
            in check_binary_comp op (fst te1) , Ebincomp(op, te1, te2)
        
        (* ternary operator is a ? b : c => a should be bool, b and c      *)
        (* should have compatible types                                    *)
        | Ast.Eternary (op, e1, e2, e3) -> let te1 = type_e e1
            in let t1 = fst te1
            in if (t1 <> `boolean)
              then raise (Type_error(Clash(t1, `boolean)))
              else
                begin
                  let e2e3 = type_2_exp e2 e3
                  in let te2 = fst e2e3 and te3 = snd e2e3
                    (* TODO : check java rules... *)
                    in fst te2, Eternary(op, te1, te2, te3)
                end
        
        | Ast.Eaffect(aff, e1) -> let exp = type_e e1 in type_affect aff exp
        | Ast.Efield_access(e, str) -> type_field_access e str
        | Ast.Evar str -> type_var str
        | Ast.Ecall(str, el) -> type_call str el
        | Ast.Enew_object(ct, el) -> type_new ct el
        | Ast.Eexpr_call(exp, met, params) -> type_exp_call exp met params
        | Ast.Esuper -> type_super ()
        | Ast.Ethis -> type_this ()
        | Ast.Enull -> `nil, Enull
        | Ast.Einstanceof (e, t) -> `boolean, Einstanceof(type_e e, t)
        | Ast.Etype_cast (jt, e) -> type_cast jt e
        | Ast.Eexpr_cast (Ast.Evar(name), e2) -> type_exp_cast name e2
        | Ast.Eexpr_cast _ | Ast.Earray_access _
        | Ast.Enew_array _
        -> raise(Type_error(UnmanagedExpression(e)))
      end
  in type_e e

and get_class_constructor env c params =
  let typed_params = List.map (type_exp env) params
  and cons_list = List.map cons_2_typelist (c.ic_cons)
  in
  (* Printf.printf " " ; Printer.print_expression stdout                   *)
  (* (Ast.Enew_object([c.ic_name], params)) ; Printf.printf "\n%!" ;       *)
  let (_, cons_id, _, _, _) =
    overload type_options.verbose type_options.overloader c env.env_classes c.ic_name (List.map fst typed_params) cons_list
  in
  (* Printf.printf " => %s %i\n\n%!" c.ic_name cons_id ; *)
    cons_id, typed_params;;

let type_exp_option env exp =
  match exp with
  | None -> None
  | Some (e) -> Some (type_exp env e);;

let type_switch_label_list env l =
  let type_switch_label sl =
    match sl with
    | Ast.Scase(e) -> Scase(type_exp env e)
    | Ast.Sdefault -> Sdefault
  in List.map type_switch_label l;;

let rec type_init_exp env ie =
  match ie with
  | Ast.Simple_init(e) -> Simple_init(type_exp env e)
  | Ast.Array_init(l) -> Array_init(List.map (type_init_exp env) l);;

let rec is_type_valid_init_exp ie l =
  match ie with
  | Simple_init(e) -> is_type_valid (fst e) l
  | Array_init(el) ->
      begin
        let f e = is_type_valid_init_exp e l
        in List.iter f el
      end ;;

let is_ast_val exp =
  match exp with
  | Ast.Ebool _ | Ast.Ebyte _ | Ast.Echar _ | Ast.Eshort _ | Ast.Eint _ -> true
  | Ast.Efloat _ | Ast.Edouble _ | Ast.Estring _ -> true
  | _ -> false;;

let get_local_var_id env =
  let id = !(List.hd !(env.env_block_var_id))
  in incr (List.hd !(env.env_block_var_id));
    id;;

(* CHECK FOR VARIABLE NAME COLLISIONS TODO : merge code *)
let check_var_local_collision env name =
  try
    match Hashtbl.find env.env_hash_var name with
    | Alocal (_, d, _) when d = !(env.env_block_depth)
    -> raise (Type_error(RedeclaredVariableInBlock(name)))
    | _ -> ()
  with | Not_found -> ();; (* none found, no problem... *)

let check_var_field_collision env super_id name =
  try
    match Hashtbl.find env.env_hash_var name with
    | Afield (_, sid, _) when super_id = sid -> raise (Type_error(RedeclaredVariableInClass(name)))
    | _ -> ()
  with | Not_found -> ();; (* none found, no problem... *)

let check_var_param_collision env name =
  try
    match Hashtbl.find env.env_hash_var name with
    | Aparam (_) -> raise (Type_error(RedeclaredVariableInMethod(name)))
    | _ -> ()
  with | Not_found -> ();; (* none found, no problem... *)

(* TYPE INSTRUCTIONS *)
let rec type_instr env i =
  let rec type_e = type_exp env and type_i i =
    (* Idecl of java_type * int * string * init_expr option *)
    let type_decl jt str init_exp =
      let jt = restrict_type jt
      in let add_local_var js exp =
        begin
          let block_id = !(env.env_block_depth) and var_id = get_local_var_id env
          in
            Hashtbl.add env.env_hash_var str (Alocal(jt, block_id, var_id)) ;
            Idecl(jt, str, exp)
        end
      in
      (* Check if there is a variable with the same name in this block *)
        check_var_local_collision env str ;
        match init_exp with
        | None -> add_local_var jt (Simple_init(get_type_default_value jt))
        | Some(e) ->
            begin
              let typed_ie = type_init_exp env e
              in
              (* check type validity... *)
                is_type_valid_init_exp typed_ie [ jt ];
                add_local_var jt typed_ie
            end
    
    (* Ast.Iswitch of expression * (switch_label list * block) list *)
    and type_switch e l =
      let type_switch_entry se =
        (type_switch_label_list env (fst se), List.map type_i (snd se))
      in Iswitch(type_e e, List.map type_switch_entry l)
    
    (* Ast.Ifor of for_init * expression option * expression list *        *)
    (* instruction TODO : change for_init...                               *)
    and type_for init cond iter instr =
      let expr_to_instr l =
        let make_instr e = (Ast.Iexpr(e))
        in List.map make_instr l
      and fdecls_to_idecls l =
        begin let make_instr (jt, str, ieo) = (Ast.Idecl(jt, str, ieo))
          in List.map make_instr l
        end
      in let for_init =
        match init with
        | Fnone -> []
        | Fexprs (el) -> expr_to_instr el
        | Fdecls (l) -> fdecls_to_idecls l
      and cond_exp =
        match cond with
        | None -> Ast.Ebool(true)
        | Some e -> e
      and iter_instr =
        let make_instr e = (Ast.Iexpr(e))
        in List.map make_instr iter
      in
      (* transform into a while block and type... *)
        type_i (Ast.Iblock(
              for_init @ [ Ast.Iwhile(cond_exp, Ast.Iblock([ instr ] @ iter_instr)) ]))
    
    and type_try bloc l bo =
      let typed_block = type_i (Ast.Iblock(bloc))
      and typed_block_option =
        match bo with
        | None -> None
        | Some (b) -> Some(type_i (Ast.Iblock(b)))
      and typed_list =
        let type_list_item (t, s, b) = (t, s, type_i (Ast.Iblock(b)))
        in List.map type_list_item l
      in
        Itry(typed_block, typed_list, typed_block_option)
    
    and type_block l =
      begin
        (* Generate 2 lists : declarations and instructions *)
        let decl_list = ref []
        and instr_list = ref []
        in let parse_instr i =
          match i with
          | Ast.Idecl (jt, str, init_opt)
          (* remove declaration, and transform into affectation if needed *)
          -> let init =
                begin
                  match init_opt with
                  | None -> None
                  | Some(Ast.Simple_init e) ->
                      begin
                        if (is_ast_val e) then Some(Ast.Simple_init(e))
                        (* cmplx exp, do default affect and evaluate later *)
                        else
                          begin
                            let new_instr = Ast.Iexpr(Ast.Eaffect (Ast.Aname([ str ]), e))
                            in instr_list := new_instr::!instr_list ;
                              None
                          end
                      end
                  | _ -> raise (Failure("Array inits are evil...\n"))
                end
              in decl_list := Ast.Idecl(jt, str, init)::!decl_list ;
          | _ -> instr_list := i::!instr_list
        in List.iter parse_instr l ;
          begin
            incr env.env_block_depth;
            env.env_block_var_id := (ref 0)::!(env.env_block_var_id);
            
            let decls = List.map type_i (List.rev !decl_list)
            and instrs = List.map type_i (List.rev !instr_list)
            in let ib = Iblock(decls, instrs)
            in
              decr env.env_block_depth ;
              env.env_block_var_id := List.tl !(env.env_block_var_id);
              (* Remove local variables from env... *)
              let rem_var v =
                match v with
                | Idecl(_, str, _) -> Hashtbl.remove env.env_hash_var str
                | _ -> ()
              in List.iter rem_var decls ;
                ib
          end
      end
    
    and type_return eo =
      let return_type = match env.env_current_method.method_return with
        | None -> `void
        | Some (t) -> restrict_type t
      in let valid_return = match eo with
        | None -> if (return_type = `void) then None
            else raise (Type_error(InvalidReturnType(`void, return_type)))
        | Some(e) ->
            begin
              let ret = type_e e
              in try
                  Some(implicit_cast env ret return_type)
                with
                | Type_error(InvalidImplicitCast(t1, t2)) -> raise (Type_error(InvalidReturnType(t1, t2)))
            end
      in Ireturn(valid_return)
    
    and type_ifthenelse cond i1 i2 =
      let c = implicit_cast env (type_e cond) `boolean
      in Iifthenelse (c, type_i i1, type_i i2)
    in
      begin match i with
        | Ast.Ilabeled (s, i) -> Ilabeled(s, type_i i)
        | Ast.Iexpr (e) -> Iexpr( type_e e)
        | Ast.Idecl (jt, str, e) -> type_decl jt str e
        | Ast.Iblock (l) -> type_block l
        | Ast.Iifthenelse (e, i1, i2) -> type_ifthenelse e i1 i2
        | Ast.Iswitch (e, l) -> type_switch e l
        
        (* check that expression has type boolean... *)
        | Ast.Iwhile (e, i) -> let te = type_e e in
              is_type_valid (fst te) [ `boolean ] ; Iwhile(type_e e, type_i i)
        | Ast.Idowhile (i, e) -> let te = type_e e in
              is_type_valid (fst te) [ `boolean ] ; Idowhile(type_i i, type_e e)
        | Ast.Ifor (init, cond, iter, instrs) -> type_for init cond iter instrs
        
        (* TODO : check validity... (inside block ?) all following         *)
        (* instructions need to be checked                                 *)
        | Ast.Ibreak (str) -> Ibreak(str)
        | Ast.Icontinue (str) -> Icontinue(str)
        (* TODO : check type... *)
        | Ast.Ireturn (eo) -> type_return eo
        | Ast.Ithrow (e) -> Ithrow(type_e e)
        | Ast.Isynchronized (e, b) -> Isynchronized(type_e e, type_i (Ast.Iblock(b)))
        | Ast.Itry (b, l, bo) -> type_try b l bo
        | Ast.Inop -> Inop
        | Ast.Iprintln (e) -> let te = implicit_cast env (type_e e) `string in Iprintln(te)
      end
  in type_i i;;

(* CLASS FIELD *)
let type_class_field super_id n {
      field_modifiers = field_modifiers;
      field_type = field_type;
      field_name = field_name;
      field_initializer = field_initializer;
    } =
  let t = restrict_type field_type and id = !n
  in incr n ;
    {
      typed_field_super_id = super_id;
      typed_field_id = id;
      typed_field_modifiers = field_modifiers;
      typed_field_type = t;
      typed_field_name = field_name;
      typed_field_initializer =
        match field_initializer with
        | Some(init) -> type_init_exp empty_env init (* TODO : check type ! *)
        | None -> Simple_init(get_type_default_value t)
    };;

let rec type_fields c =
  let n = ref 0
  in let rec tf c super_id n =
    let parent_fields = match c.ic_father with
      | None -> []
      | Some(p) -> tf p (super_id + 1) n
    in List.append parent_fields (List.map (type_class_field super_id n) c.ic_class.class_fields)
  in tf c 0 n;;

(* PARAMS *)
let add_params_to_env env params =
  let n = ref 0
  in let add_param_to_env (jt, str) =
    check_var_param_collision env str;
    Hashtbl.add env.env_hash_var str (Aparam(jt, !n)) ;
    incr n
  in List.iter add_param_to_env params;;

let rem_params_from_env env params =
  let rem_param_from_env env (_, str) =
    Hashtbl.remove env.env_hash_var str
  in List.iter (rem_param_from_env env) params;;

(* CLASS METHOD *)

let method_restrict_types {
      method_modifiers = method_modifiers;
      method_return = method_return;
      method_name = method_name;
      method_parameters = method_parameters;
      method_throws = method_throws;
      method_body = method_body;
    } = let param_restrict_types (t, s) = (restrict_type t, s)
  in {
      method_modifiers = method_modifiers;
      method_return = (match method_return with | None -> None | Some (t) -> Some(restrict_type t));
      method_name = method_name;
      method_parameters = List.map param_restrict_types method_parameters;
      method_throws = method_throws;
      method_body = method_body;
    };;

let methods_restrict_types l =
  List.map method_restrict_types l;;

let constructor_restrict_types {
      constructor_modifiers = constructor_modifiers;
      constructor_name = constructor_name;
      constructor_parameters = constructor_parameters;
      constructor_throws = constructor_throws;
      constructor_call = constructor_call;
      constructor_body = constructor_body;
    } = let param_restrict_types (t, s) = (restrict_type t, s)
  in {
      constructor_modifiers = constructor_modifiers;
      constructor_name = constructor_name;
      constructor_parameters = List.map param_restrict_types constructor_parameters;
      constructor_throws = constructor_throws;
      constructor_call = constructor_call;
      constructor_body = constructor_body;
    };;

let constructors_restrict_types l =
  List.map constructor_restrict_types l;;

let type_class_method env met =
  let mangled_name = mangle_method met.method_name met.method_parameters
  and ret_type = match met.method_return with None -> `void | Some(t) -> t
  in if type_options.verbose then Printf.printf "  type %s : %s\n%!" mangled_name (string_of_java_type ret_type);
    add_params_to_env env met.method_parameters ;
    env.env_current_method <- met;
    let body = match met.method_body with
      | None -> None
      | Some(b) -> Some(type_instr env (Ast.Iblock(b)))
    in
      rem_params_from_env env met.method_parameters ;
      {
        typed_method_modifiers = met.method_modifiers;
        typed_method_return = met.method_return;
        typed_method_name = met.method_name;
        typed_method_mangled_name = mangled_name;
        typed_method_parameters = met.method_parameters;
        typed_method_throws = met.method_throws;
        typed_method_body = body;
      }
;;

(* CLASS CONSTRUCTOR *)

let type_class_cons env cons =
  let cons = constructor_restrict_types cons
  in let c = env.env_current_class
  in add_params_to_env env cons.constructor_parameters ;
    let type_ccall cc =
      (* check existence of constructors... *)
      match cc with
      | Ast.Cnone -> Cnone
      | Ast.Cthis (el) -> let id, params = get_class_constructor env c el
          in Cthis(id, params)
      | Ast.Csuper(el) -> let id, params = get_class_constructor env (get_class_parent c) el
          in Csuper(id, params)
    and body = match cons.constructor_body with
      | None -> None
      | Some (b) -> Some(type_instr env (Ast.Iblock(b)))
    in rem_params_from_env env cons.constructor_parameters ;
      {
        typed_constructor_modifiers = cons.constructor_modifiers;
        typed_constructor_name = env.env_current_class.ic_name;
        typed_constructor_parameters = cons.constructor_parameters;
        typed_constructor_throws = cons.constructor_throws;
        typed_constructor_call = type_ccall cons.constructor_call;
        typed_constructor_body = body;
      };;

let add_field_to_env env {
      typed_field_super_id = typed_field_super_id;
      typed_field_id = typed_field_id ;
      typed_field_type = typed_field_type ;
      typed_field_name = typed_field_name ;
    } =
  
  check_var_field_collision env typed_field_super_id typed_field_name;
  Hashtbl.add env.env_hash_var typed_field_name
    (Afield(typed_field_type, typed_field_super_id, typed_field_id))
;;

let add_method_to_env env {
      method_modifiers = method_modifiers;
      method_return = method_return;
      method_name = method_name;
      method_parameters = method_parameters;
      method_throws = method_throws;
      method_body = method_body;
    } =
  let meth = {
    env_method_modifiers = method_modifiers;
    env_method_return = (match method_return with None -> None | Some(t) -> Some(restrict_type t));
    env_method_parameters = method_parameters;
    env_method_throws = method_throws;
  } and mangled_name = method_name ^ "_" ^ (string_of_int (List.length method_parameters))
  in
  (* Printf.printf " add method %s\n%!" (mangle_method method_name         *)
  (* method_parameters) ;                                                  *)
  
    Hashtbl.add env.env_hash_meth mangled_name meth;;

let rec get_class_in_tree name t =
  let is_class t =
    (String.compare name t.ic_name) = 0
  in let rec get_class_in_list l =
    begin
      match l with
      | [] -> None
      | h:: t ->
          begin
            let par = get_class_in_tree name h
            in match par with
              | None -> get_class_in_list t
              | _ -> par
          end
    end
  in
    if (is_class t)
    then Some(t)
    else get_class_in_list t.ic_children;;

let print_hierarchy obj =
  let rec print_indent n =
    if (n > 0) then
      (Printf.printf "  " ; print_indent (n - 1))
  in
  let rec print_hier n obj =
    print_indent n ;
    Printf.printf "%s\n%!" obj.ic_name ;
    List.iter (print_hier (n + 1)) obj.ic_children;
  in print_hier 0 obj;;

(* CLASS *)
let type_default_constructor env class_name has_super =
  [ type_class_cons env (create_default_constructor class_name has_super) ];;

let type_class obj c {
      class_modifiers = class_modifiers;
      class_name = class_name;
      class_extend = class_extend;
      class_implements = class_implements;
      class_fields = class_fields;
      class_methods = class_methods;
      class_initializers = class_initializers;
      class_constructors = class_constructors;
    } =
  (* Printf.printf "\ntype_class %s ...\n%!" class_name ; *)
  let methods = methods_restrict_types class_methods
  and fields = c.ic_fields
  in let env = {
    env_current_class = c;
    env_current_method = empty_method;
    env_field_decls = fields; (* typed fields *)
    env_method_def = methods; (* method list... *)
    env_hash_var = Hashtbl.create 256;
    env_hash_meth = Hashtbl.create 256;
    env_block_depth = ref (- 1);
    env_block_var_id = ref [];
    env_classes = obj;
  }
  in
    List.iter (add_field_to_env env) fields ;
    List.iter (add_method_to_env env) methods ;
    let class_init =
      let instr_from_block b = Ast.Iblock(b)
      in List.map (type_instr env) (List.map instr_from_block class_initializers)
    and constructors = match class_constructors with
      | [] -> type_default_constructor env class_name (c.ic_father != None)
      | _ -> List.map (type_class_cons env) class_constructors
    in {
        typed_class_modifiers = class_modifiers;
        typed_class_name = class_name;
        typed_class_extend = class_extend;
        typed_class_implements = class_implements;
        typed_class_fields = fields;
        typed_class_methods = List.map (type_class_method env) methods;
        typed_class_initializers = class_init ;
        typed_class_constructors = constructors;
      };;

let rec list_merge l1 l2 cmp =
  match l1 with
  | [] -> l2
  | ((i1, i2, str, m) as h):: t ->
      begin
        if (List.exists (cmp h) l2) then list_merge t l2 cmp (* ignore *)
        else ((i1 + 1), i2, str, m):: (list_merge t l2 cmp)
      end;;

let method_2_name m =
  let get_type (t, _) = t
  in let rec name_param l =
    match l with
    | [] -> ""
    | h::[] -> string_of_java_type (get_type h) ^ " "
    | h:: t -> (string_of_java_type (get_type h)) ^ ", " ^ name_param t
  in
    m.method_name ^ "( " ^ (name_param m.method_parameters) ^ ")";;

let methods_compare (_, _, str1, _) (_, _, str2, _) =
  (String.compare str1 str2) = 0;;

let method_2_id list_parent n m =
  let m = method_restrict_types m
  in let name = method_2_name m
  in try
      let (_, id, _, _) = List.find (methods_compare (0, 0, name, m)) list_parent
      in
        (0, id, name, m)  (* reuse method id... *)
    with | Not_found ->
    (* add new method ! *)
        let item = (0, !n, name, m) in incr n ; item;;

let cons_2_id n cons =
  let id = !n
  in incr n ; (id, cons);;

let rec add_classes_to_tree classes obj =
  let insert_node c p =
    (* Compare 2 methods (father and son methods) *)
    let n = ref (List.length p.ic_methods) and n_cons = ref 0
    in let child_methods = List.map (method_2_id p.ic_methods n) c.class_methods
    in let methods = list_merge p.ic_methods child_methods methods_compare
    and cons = match c.class_constructors with
      | [] -> [ (0, create_default_constructor c.class_name true) ]
      | _ -> List.map (cons_2_id n_cons) c.class_constructors
    in let child = {
      ic_name = c.class_name;
      ic_abstract = List.exists (fun x -> (x = `abstract)) c.class_modifiers;
      ic_father = Some(p);
      ic_children = [];
      ic_class = c;
      ic_fields = [];
      ic_methods = methods;
      ic_cons = cons
    }
    in begin
        let is_abstract (_, _, _, m) = is_method_abstract m
        in if (not child.ic_abstract) && (List.exists is_abstract child.ic_methods)
          then raise (Type_error(AbstractMethodInClass(child.ic_name)));
      end ;
      p.ic_children <- child:: p.ic_children;
      child.ic_fields <- type_fields child
  in match classes with
    | [] -> ()
    | h:: rest ->
        begin
          match h.class_extend with
          | [] -> insert_node h obj ; (* parent is object, check the rest *)
              add_classes_to_tree rest obj
          | super:: trash ->
              begin
                let parent = get_class_in_tree super obj
                in match parent with
                  | Some(p) ->
                      begin
                        insert_node h p ;
                        add_classes_to_tree rest obj
                      end
                  | None ->
                  
                      let is_parent p = (String.compare super p.class_name) = 0
                      in
                        (
                          try let _ = List.find is_parent rest in ()
                          with | Not_found -> raise (Type_error(UndeclaredClass(super)))
                        );
                        let new_list = rest @ [h] in add_classes_to_tree new_list obj
              end
        end;;

let gen_class_hierarchy classes =
  (* Create father... *)
  let obj = {
    ic_name = "Object";
    ic_abstract = false;
    ic_father = None;
    ic_children = [];
    ic_class = obj_class;
    ic_fields = [];
    ic_methods = [];
    ic_cons = [ (0, create_default_constructor "Object" false) ];
  } in
    add_classes_to_tree classes obj;
    if type_options.verbose then print_hierarchy obj ;
    obj
;;

let methods_2_array typed_methods ic_methods parent_methods =
  let n = List.length ic_methods
  in let a = Array.make n (0, 0, "", blank_typed_method)
  in let process_method typed_methods parent_methods ic_method =
    match ic_method with
    | (0, method_id, name, m) ->
        begin
          (* Method should be present in this class... *)
          let cmp name tm =
            (String.compare name tm.typed_method_mangled_name) = 0
          in let mangled_name = (mangle_method m.method_name m.method_parameters)
          in let tm = List.find (cmp mangled_name) typed_methods
          in a.(method_id) <- (0, method_id, name, tm)
        end
    | (parent_id, method_id, name, m) ->
    (* Get method in parent class *)
        let _, _, _, typed_method = parent_methods.(method_id)
        in a.(method_id) <- (parent_id, method_id, name, typed_method)
  (* parcourir la liste des methodes, chercher dans les methodes typees ou *)
  (* prend chez le parent                                                  *)
  in List.iter (process_method typed_methods parent_methods) ic_methods;
    a;;

let cons_2_array constructors =
  Array.of_list constructors

let type_class_hier obj =
  let rec rec_class_hier obj parent ic_c =
    let {
      ic_name = ic_name;
      ic_father = ic_father;
      ic_children = ic_children;
      ic_class = ic_class;
      ic_fields = ic_fields;
      ic_methods = ic_methods;
      ic_cons = ic_cons;
    } = ic_c
    in if type_options.verbose then Printf.printf "Type %s\n%!" ic_name ;
      begin
        let tc = type_class obj ic_c ic_class
        in let fields = tc.typed_class_fields
        and m_count = List.length ic_methods
        in let methods =
          let parent_methods =
            match parent with
            | None -> [||]
            | Some (p) -> p.tc_methods
          in methods_2_array tc.typed_class_methods ic_methods parent_methods;
        
        and constructors = cons_2_array tc.typed_class_constructors
        in let new_class =
          {
            tc_name = ic_name;
            tc_father = parent;
            tc_children = [];
            tc_class = tc;
            tc_field_count = List.length fields;
            tc_fields = fields;
            tc_method_count = m_count;
            tc_methods = methods;
            tc_cons = constructors;
          } in
          begin
            (* add to parent... *)
            begin
              match parent with
              | None -> ()
              | Some(p) -> p.tc_children <- new_class:: p.tc_children
            end ;
            List.iter (fun x -> ignore (rec_class_hier obj (Some(new_class)) x)) ic_children
          end ;
          new_class
      end
  in rec_class_hier obj None obj
;;

(* COMPILE UNIT *)
let type_compil_unit {
      package = package;
      imports = imports;
      class_decls = class_decls;
      interf_decls = interf_decls;
    } verbose overloader =
  type_options.verbose <- verbose;
  type_options.overloader <- overloader;
  let obj = gen_class_hierarchy class_decls in
    {
      typed_package = package;
      typed_imports = imports;
      typed_class = type_class_hier obj;
      typed_interf_decls = interf_decls;
    };;

